@@ -25,6 +25,7 @@ except ModuleNotFoundError:
     import re  # type: ignore
 
 import flask
+import werkzeug.exceptions as exceptions
 import ProxyPatternPool as ppp  # type: ignore
 
 # for local use & forwarding
@@ -41,6 +42,7 @@ import logging
 log = logging.getLogger("fsa")
 
 # get module version
+# TODO update when 3.7 is not supported anymore
 import pkg_resources as pkg  # type: ignore
 __version__ = pkg.require("FlaskSimpleAuth")[0].version
 
@@ -61,6 +63,7 @@ AfterRequestFun = Callable[[Response], Response]
 @dataclass
 class ErrorResponse(BaseException):
     """Internal exception class to carry fields for an error Response."""
+    # NOTE this should maybe inherit from exceptions.HTTPException?
 
     message: str
     status: int
@@ -232,6 +235,7 @@ _DIRECTIVES = {
     "FSA_MODE", "FSA_LOGGING_LEVEL",
     # general settings
     "FSA_SECURE", "FSA_SERVER_ERROR", "FSA_NOT_FOUND_ERROR", "FSA_LOCAL",
+    "FSA_HANDLE_ALL_ERRORS",
     # register hooks
     "FSA_GET_USER_PASS", "FSA_USER_IN_GROUP", "FSA_CAST",
     "FSA_OBJECT_PERMS", "FSA_SPECIAL_PARAMETER", "FSA_ERROR_RESPONSE",
@@ -308,9 +312,11 @@ class FlaskSimpleAuth:
         self._pm = None
         self._cache: Optional[MutableMapping[str, str]] = None
         self._gen_cache: Optional[Callable] = None
+        # fsa-generated errors
         self._server_error: int = _DEFAULT_SERVER_ERROR
         self._not_found_error: int = _DEFAULT_NOT_FOUND_ERROR
         self._secure: bool = True
+        self._secure_warning = True
         self._names: Set[str] = set()
         self._groups: Set[Union[str, int]] = set()
         self._scopes: Set[str] = set()
@@ -368,12 +374,13 @@ class FlaskSimpleAuth:
         if not request.is_secure and not (
             request.remote_addr.startswith("127.") or request.remote_addr == "::1"
         ):  # pragma: no cover
-            msg = f"insecure HTTP request on {request.remote_addr}, allow with FSA_SECURE=False"
             if self._secure:
-                log.error(msg)
+                log.error("insecure HTTP request, allow with FSA_SECURE=False")
                 return self._Res("insecure HTTP request denied", self._server_error)
-            else:  # at least a warning is issued for each insecure request
-                log.warning(msg)
+            else:  # one warning is issued
+                if self._secure_warning:
+                    log.warning("insecure HTTP request seen")
+                    self._secure_warning = False
 
     def _auth_reset_user(self):
         """Before request hook to cleanup authentication and authorization."""
@@ -468,6 +475,7 @@ class FlaskSimpleAuth:
         return res
 
     def _add_delay_header(self, res: Response):
+        """Add delay showing elapsed time from module perspective."""
         delay = dt.datetime.timestamp(dt.datetime.now()) - self._local.start
         res.headers["FSA-Delay"] = f"{delay:.6f}"
         return res
@@ -628,8 +636,6 @@ class FlaskSimpleAuth:
         # status code for some errors errors
         self._server_error = conf.get("FSA_SERVER_ERROR", _DEFAULT_SERVER_ERROR)
         self._not_found_error = conf.get("FSA_NOT_FOUND_ERROR", _DEFAULT_NOT_FOUND_ERROR)
-        # whether to error on unexpected parameters
-        self._reject_param = conf.get("FSA_REJECT_UNEXPECTED_PARAM", _DEFAULT_REJECT_UNEXPECTED_PARAM)
         # actual error response generation
         if self._error_response is None:
             error = conf.get("FSA_ERROR_RESPONSE", _DEFAULT_ERROR_RESPONSE)
@@ -642,14 +648,20 @@ class FlaskSimpleAuth:
             elif error == "plain":
                 self._error_response = lambda m, c: Response(m, c, content_type="text/plain")
             elif error == "json":
-                self._error_response = lambda m, c: Response(json.dumps(m), c, content_type="text/json")
+                self._error_response = lambda m, c: Response(json.dumps(m), c, content_type="application/json")
             elif error.startswith("json:"):
                 key = error.split(":", 1)[1]
-                self._error_response = lambda m, c: Response(json.dumps({key: m}), c, content_type="text/json")
+                self._error_response = lambda m, c: Response(json.dumps({key: m}), c, content_type="application/json")
             if self._error_response is None:
                 raise self._Bad(f"unexpected FSA_ERROR_RESPONSE value: {error}")
         elif "FSA_ERROR_RESPONSE" in conf:
             log.warning("ignoring FSA_ERROR_RESPONSE directive, handler already set")
+        # possibly take responsability for handling errors
+        if conf.get("FSA_HANDLE_ALL_ERRORS", True):
+            self._app.register_error_handler(exceptions.HTTPException,
+                                             lambda e: self._Res(e.description, e.code))
+        # whether to error on unexpected parameters
+        self._reject_param = conf.get("FSA_REJECT_UNEXPECTED_PARAM", _DEFAULT_REJECT_UNEXPECTED_PARAM)
         #
         # overall authn setup
         #
@@ -1463,7 +1475,7 @@ class FlaskSimpleAuth:
 
                 for scope in scopes:
                     if not self.user_scope(scope):
-                        return self._Res("", 403)
+                        return self._Res(f"missing permission \"{scope}\"", 403)
 
                 return self._safe_call(path, "oauth authorization", fun, *args, **kwargs)
 
@@ -1507,7 +1519,7 @@ class FlaskSimpleAuth:
                         log.error(f"type error in user_in_group: {ok}: {type(ok)}, must return a boolean")
                         return self._Res("internal error with user_in_group", self._server_error)
                     if not ok:
-                        return self._Res("", 403)
+                        return self._Res(f"not in group \"{grp}\"", 403)
 
                 # all groups are ok, proceed to call underlying function
                 return self._safe_call(path, "group authorization", fun, *args, **kwargs)
@@ -1668,7 +1680,7 @@ class FlaskSimpleAuth:
                         log.error(f"type error on on {request.method} {request.path} permission {perm} check: {type(ok)}")
                         return self._Res("internal error with permission check", self._server_error)
                     elif not ok:
-                        return self._Res("", 403)
+                        return self._Res(f"permission denied on {domain}/{val} {mode}", 403)
                     # else: all is well, check next!
 
                 # then call the initial function
